/*     */ package org.apache.avro.file;
/*     */ 
/*     */ import java.io.Closeable;
/*     */ import java.io.EOFException;
/*     */ import java.io.IOException;
/*     */ import java.io.InputStream;
/*     */ import java.io.UnsupportedEncodingException;
/*     */ import java.nio.ByteBuffer;
/*     */ import java.util.ArrayList;
/*     */ import java.util.Arrays;
/*     */ import java.util.Collections;
/*     */ import java.util.HashMap;
/*     */ import java.util.Iterator;
/*     */ import java.util.List;
/*     */ import java.util.Map;
/*     */ import java.util.NoSuchElementException;
/*     */ import org.apache.avro.AvroRuntimeException;
/*     */ import org.apache.avro.Schema;
/*     */ import org.apache.avro.io.BinaryDecoder;
/*     */ import org.apache.avro.io.BinaryEncoder;
/*     */ import org.apache.avro.io.DatumReader;
/*     */ import org.apache.avro.io.DecoderFactory;
/*     */ import org.apache.avro.util.Utf8;
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/*     */ public class DataFileStream<D>
/*     */   implements Iterator<D>, Iterable<D>, Closeable
/*     */ {
/*     */   private DatumReader<D> reader;
/*     */   private long blockSize;
/*     */   
/*     */   public static final class Header
/*     */   {
/*     */     Schema schema;
/*  54 */     Map<String, byte[]> meta = new HashMap();
/*  55 */     private transient List<String> metaKeyList = new ArrayList();
/*  56 */     byte[] sync = new byte[16];
/*     */   }
/*     */   
/*     */ 
/*     */ 
/*     */ 
/*  62 */   private boolean availableBlock = false;
/*     */   
/*     */ 
/*     */   private Header header;
/*     */   
/*     */   BinaryDecoder vin;
/*     */   
/*  69 */   BinaryDecoder datumIn = null;
/*     */   
/*     */   ByteBuffer blockBuffer;
/*     */   long blockCount;
/*     */   long blockRemaining;
/*  74 */   byte[] syncBuffer = new byte[16];
/*     */   
/*     */ 
/*     */   private Codec codec;
/*     */   
/*     */ 
/*     */   public DataFileStream(InputStream in, DatumReader<D> reader)
/*     */     throws IOException
/*     */   {
/*  83 */     this.reader = reader;
/*  84 */     initialize(in);
/*     */   }
/*     */   
/*     */ 
/*     */   protected DataFileStream(DatumReader<D> reader)
/*     */     throws IOException
/*     */   {
/*  91 */     this.reader = reader;
/*     */   }
/*     */   
/*     */   void initialize(InputStream in) throws IOException
/*     */   {
/*  96 */     this.header = new Header(null);
/*  97 */     this.vin = DecoderFactory.get().binaryDecoder(in, this.vin);
/*  98 */     byte[] magic = new byte[DataFileConstants.MAGIC.length];
/*     */     try {
/* 100 */       this.vin.readFixed(magic);
/*     */     } catch (IOException e) {
/* 102 */       throw new IOException("Not a data file.");
/*     */     }
/* 104 */     if (!Arrays.equals(DataFileConstants.MAGIC, magic)) {
/* 105 */       throw new IOException("Not a data file.");
/*     */     }
/* 107 */     long l = this.vin.readMapStart();
/* 108 */     if (l > 0L) {
/*     */       do {
/* 110 */         for (long i = 0L; i < l; i += 1L) {
/* 111 */           String key = this.vin.readString(null).toString();
/* 112 */           ByteBuffer value = this.vin.readBytes(null);
/* 113 */           byte[] bb = new byte[value.remaining()];
/* 114 */           value.get(bb);
/* 115 */           this.header.meta.put(key, bb);
/* 116 */           this.header.metaKeyList.add(key);
/*     */         }
/* 118 */       } while ((l = this.vin.mapNext()) != 0L);
/*     */     }
/* 120 */     this.vin.readFixed(this.header.sync);
/*     */     
/*     */ 
/* 123 */     this.header.metaKeyList = Collections.unmodifiableList(this.header.metaKeyList);
/* 124 */     this.header.schema = Schema.parse(getMetaString("avro.schema"), false);
/* 125 */     this.codec = resolveCodec();
/* 126 */     this.reader.setSchema(this.header.schema);
/*     */   }
/*     */   
/*     */   void initialize(InputStream in, Header header) throws IOException
/*     */   {
/* 131 */     this.header = header;
/* 132 */     this.codec = resolveCodec();
/* 133 */     this.reader.setSchema(header.schema);
/*     */   }
/*     */   
/*     */   Codec resolveCodec() {
/* 137 */     String codecStr = getMetaString("avro.codec");
/* 138 */     if (codecStr != null) {
/* 139 */       return CodecFactory.fromString(codecStr).createInstance();
/*     */     }
/* 141 */     return CodecFactory.nullCodec().createInstance();
/*     */   }
/*     */   
/*     */ 
/*     */   public Header getHeader()
/*     */   {
/* 147 */     return this.header;
/*     */   }
/*     */   
/* 150 */   public Schema getSchema() { return this.header.schema; }
/*     */   
/*     */   public List<String> getMetaKeys()
/*     */   {
/* 154 */     return this.header.metaKeyList;
/*     */   }
/*     */   
/*     */   public byte[] getMeta(String key)
/*     */   {
/* 159 */     return (byte[])this.header.meta.get(key);
/*     */   }
/*     */   
/*     */   public String getMetaString(String key) {
/* 163 */     byte[] value = getMeta(key);
/* 164 */     if (value == null) {
/* 165 */       return null;
/*     */     }
/*     */     try {
/* 168 */       return new String(value, "UTF-8");
/*     */     } catch (UnsupportedEncodingException e) {
/* 170 */       throw new RuntimeException(e);
/*     */     }
/*     */   }
/*     */   
/*     */   public long getMetaLong(String key) {
/* 175 */     return Long.parseLong(getMetaString(key));
/*     */   }
/*     */   
/*     */ 
/*     */ 
/*     */ 
/*     */ 
/* 182 */   public Iterator<D> iterator() { return this; }
/*     */   
/* 184 */   private DataBlock block = null;
/*     */   
/*     */   public boolean hasNext()
/*     */   {
/*     */     try {
/* 189 */       if (this.blockRemaining == 0L)
/*     */       {
/* 191 */         if (null != this.datumIn) {
/* 192 */           boolean atEnd = this.datumIn.isEnd();
/* 193 */           if (!atEnd) {
/* 194 */             throw new IOException("Block read partially, the data may be corrupt");
/*     */           }
/*     */         }
/* 197 */         if (hasNextBlock()) {
/* 198 */           this.block = nextRawBlock(this.block);
/* 199 */           this.block.decompressUsing(this.codec);
/* 200 */           this.blockBuffer = this.block.getAsByteBuffer();
/* 201 */           this.datumIn = DecoderFactory.get().binaryDecoder(this.blockBuffer.array(), this.blockBuffer.arrayOffset() + this.blockBuffer.position(), this.blockBuffer.remaining(), this.datumIn);
/*     */         }
/*     */       }
/*     */       
/*     */ 
/* 206 */       return this.blockRemaining != 0L;
/*     */     } catch (EOFException e) {
/* 208 */       return false;
/*     */     } catch (IOException e) {
/* 210 */       throw new AvroRuntimeException(e);
/*     */     }
/*     */   }
/*     */   
/*     */ 
/*     */ 
/*     */   public D next()
/*     */   {
/*     */     try
/*     */     {
/* 220 */       return (D)next(null);
/*     */     } catch (IOException e) {
/* 222 */       throw new AvroRuntimeException(e);
/*     */     }
/*     */   }
/*     */   
/*     */ 
/*     */ 
/*     */   public D next(D reuse)
/*     */     throws IOException
/*     */   {
/* 231 */     if (!hasNext())
/* 232 */       throw new NoSuchElementException();
/* 233 */     D result = this.reader.read(reuse, this.datumIn);
/* 234 */     if (0L == --this.blockRemaining) {
/* 235 */       blockFinished();
/*     */     }
/* 237 */     return result;
/*     */   }
/*     */   
/*     */   public ByteBuffer nextBlock() throws IOException
/*     */   {
/* 242 */     if (!hasNext())
/* 243 */       throw new NoSuchElementException();
/* 244 */     if (this.blockRemaining != this.blockCount)
/* 245 */       throw new IllegalStateException("Not at block start.");
/* 246 */     this.blockRemaining = 0L;
/* 247 */     this.datumIn = null;
/* 248 */     return this.blockBuffer;
/*     */   }
/*     */   
/*     */   public long getBlockCount() {
/* 252 */     return this.blockCount;
/*     */   }
/*     */   
/*     */   protected void blockFinished() throws IOException
/*     */   {}
/*     */   
/*     */   boolean hasNextBlock() {
/*     */     try {
/* 260 */       if (this.availableBlock) return true;
/* 261 */       if (this.vin.isEnd()) return false;
/* 262 */       this.blockRemaining = this.vin.readLong();
/* 263 */       this.blockSize = this.vin.readLong();
/* 264 */       if ((this.blockSize > 2147483647L) || (this.blockSize < 0L))
/*     */       {
/* 266 */         throw new IOException("Block size invalid or too large for this implementation: " + this.blockSize);
/*     */       }
/*     */       
/* 269 */       this.blockCount = this.blockRemaining;
/* 270 */       this.availableBlock = true;
/* 271 */       return true;
/*     */     } catch (EOFException eof) {
/* 273 */       return false;
/*     */     } catch (IOException e) {
/* 275 */       throw new AvroRuntimeException(e);
/*     */     }
/*     */   }
/*     */   
/*     */   DataBlock nextRawBlock(DataBlock reuse) throws IOException {
/* 280 */     if (!hasNextBlock()) {
/* 281 */       throw new NoSuchElementException();
/*     */     }
/* 283 */     if ((reuse == null) || (reuse.data.length < (int)this.blockSize)) {
/* 284 */       reuse = new DataBlock(this.blockRemaining, (int)this.blockSize, null);
/*     */     } else {
/* 286 */       reuse.numEntries = this.blockRemaining;
/* 287 */       reuse.blockSize = ((int)this.blockSize);
/*     */     }
/*     */     
/* 290 */     this.vin.readFixed(reuse.data, 0, reuse.blockSize);
/* 291 */     this.vin.readFixed(this.syncBuffer);
/* 292 */     if (!Arrays.equals(this.syncBuffer, this.header.sync))
/* 293 */       throw new IOException("Invalid sync!");
/* 294 */     this.availableBlock = false;
/* 295 */     return reuse;
/*     */   }
/*     */   
/*     */   public void remove()
/*     */   {
/* 300 */     throw new UnsupportedOperationException();
/*     */   }
/*     */   
/*     */   public void close() throws IOException
/*     */   {
/* 305 */     this.vin.inputStream().close();
/*     */   }
/*     */   
/*     */   static class DataBlock {
/*     */     private byte[] data;
/*     */     private long numEntries;
/*     */     private int blockSize;
/* 312 */     private int offset = 0;
/*     */     
/* 314 */     private DataBlock(long numEntries, int blockSize) { this.data = new byte[blockSize];
/* 315 */       this.numEntries = numEntries;
/* 316 */       this.blockSize = blockSize;
/*     */     }
/*     */     
/*     */     DataBlock(ByteBuffer block, long numEntries) {
/* 320 */       this.data = block.array();
/* 321 */       this.blockSize = block.remaining();
/* 322 */       this.offset = (block.arrayOffset() + block.position());
/* 323 */       this.numEntries = numEntries;
/*     */     }
/*     */     
/*     */     byte[] getData() {
/* 327 */       return this.data;
/*     */     }
/*     */     
/*     */     long getNumEntries() {
/* 331 */       return this.numEntries;
/*     */     }
/*     */     
/*     */     int getBlockSize() {
/* 335 */       return this.blockSize;
/*     */     }
/*     */     
/*     */     ByteBuffer getAsByteBuffer() {
/* 339 */       return ByteBuffer.wrap(this.data, this.offset, this.blockSize);
/*     */     }
/*     */     
/*     */     void decompressUsing(Codec c) throws IOException {
/* 343 */       ByteBuffer result = c.decompress(getAsByteBuffer());
/* 344 */       this.data = result.array();
/* 345 */       this.blockSize = result.remaining();
/*     */     }
/*     */     
/*     */     void compressUsing(Codec c) throws IOException {
/* 349 */       ByteBuffer result = c.compress(getAsByteBuffer());
/* 350 */       this.data = result.array();
/* 351 */       this.blockSize = result.remaining();
/*     */     }
/*     */     
/*     */     void writeBlockTo(BinaryEncoder e, byte[] sync) throws IOException {
/* 355 */       e.writeLong(this.numEntries);
/* 356 */       e.writeLong(this.blockSize);
/* 357 */       e.writeFixed(this.data, this.offset, this.blockSize);
/* 358 */       e.writeFixed(sync);
/* 359 */       e.flush();
/*     */     }
/*     */   }
/*     */ }


/* Location:              C:\Perforce\MyAsup\dev\thirdparty\lib\asup-metamodel-client-1.0.jar!\org\apache\avro\file\DataFileStream.class
 * Java compiler version: 6 (50.0)
 * JD-Core Version:       0.7.1
 */